import argparse
import json
import logging
import os
import sys
import typing
from argparse import ArgumentParser

import yaml

from knowit import (
    __url__,
    __version__,
    api,
)
from knowit.provider import ProviderError
from knowit.serializer import (
    get_json_encoder,
    get_yaml_dumper,
)
from knowit.utils import recurse_paths

logging.basicConfig(stream=sys.stdout, format='%(message)s')
logging.getLogger('CONSOLE').setLevel(logging.INFO)
logging.getLogger('knowit').setLevel(logging.ERROR)

console = logging.getLogger('CONSOLE')
logger = logging.getLogger('knowit')


def build_argument_parser() -> ArgumentParser:
    """Build the argument parser."""
    opts = ArgumentParser()
    opts.add_argument(
        dest='videopath',
        help='Path to the video to introspect',
        nargs='*',
        type=str,
    )

    provider_opts = opts.add_argument_group('Providers')
    provider_opts.add_argument(
        '-p',
        '--provider',
        dest='provider',
        help='The provider to be used: mediainfo, ffmpeg, mkvmerge or enzyme.',
        type=str,
    )

    output_opts = opts.add_argument_group('Output')
    output_opts.add_argument(
        '--debug',
        action='store_true',
        dest='debug',
        help='Print information for debugging knowit and for reporting bugs.'
    )
    output_opts.add_argument(
        '--report',
        action='store_true',
        dest='report',
        help='Parse media and report all non-detected values'
    )
    output_opts.add_argument(
        '-y',
        '--yaml',
        action='store_true',
        dest='yaml',
        help='Display output in yaml format'
    )
    output_opts.add_argument(
        '-N',
        '--no-units',
        action='store_true',
        dest='no_units',
        help='Display output without units'
    )
    output_opts.add_argument(
        '-P',
        '--profile',
        dest='profile',
        help='Display values according to specified profile: code, default, human, technical',
        type=str,
    )

    conf_opts = opts.add_argument_group('Configuration')
    conf_opts.add_argument(
        '--mediainfo',
        dest='mediainfo',
        help='The location to search for MediaInfo binaries',
        type=str,
    )
    conf_opts.add_argument(
        '--ffmpeg',
        dest='ffmpeg',
        help='The location to search for ffprobe (FFmpeg) binaries',
        type=str,
    )
    conf_opts.add_argument(
        '--mkvmerge',
        dest='mkvmerge',
        help='The location to search for mkvmerge (MKVToolNix) binaries',
        type=str,
    )

    information_opts = opts.add_argument_group('Information')
    information_opts.add_argument(
        '--version',
        dest='version',
        action='store_true',
        help='Display knowit version.'
    )

    return opts


def knowit(
        video_path: typing.Union[str, os.PathLike],
        options: argparse.Namespace,
        context: typing.MutableMapping,
) -> typing.Mapping:
    """Extract video metadata."""
    context['path'] = video_path
    if not options.report:
        console.info('For: %s', video_path)
    else:
        console.info('Parsing: %s', video_path)
    info = api.know(video_path, context)
    if not options.report:
        console.info('Knowit %s found: ', __version__)
        console.info(dumps(info, options, context))
    return info


def _as_yaml(
        info: typing.Mapping[str, typing.Any],
        context: typing.Mapping,
) -> str:
    """Convert info to string using YAML format."""
    data = {info['path']: info} if 'path' in info else info
    return yaml.dump(
        data,
        Dumper=get_yaml_dumper(context),
        default_flow_style=False,
        allow_unicode=True,
        sort_keys=False,
    )


def _as_json(
        info: typing.Mapping[str, typing.Any],
        context: typing.Mapping,
) -> str:
    """Convert info to string using JSON format."""
    return json.dumps(
        info,
        cls=get_json_encoder(context),
        indent=4,
        ensure_ascii=False,
    )


def dumps(
        info: typing.Mapping[str, typing.Any],
        options: argparse.Namespace,
        context: typing.Mapping,
) -> str:
    """Convert info to string using json or yaml format."""
    convert = _as_yaml if options.yaml else _as_json
    return convert(info, context)


def main(args: typing.Optional[typing.List[str]] = None) -> None:
    """Execute main function for entry point."""
    argument_parser = build_argument_parser()
    args = args or sys.argv[1:]
    options = argument_parser.parse_args(args)

    if options.debug:
        logger.setLevel(logging.DEBUG)
        logging.getLogger('enzyme').setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    paths = recurse_paths(options.videopath)

    if not paths:
        if options.version:
            console.info(api.debug_info())
        else:
            argument_parser.print_help()
        return

    report: typing.MutableMapping[str, str] = {}
    for i, video_path in enumerate(paths):
        try:
            context = {k: v for k, v in vars(options).items() if v is not None}
            if options.report:
                context['report'] = report
            else:
                del context['report']
            knowit(video_path, options, context)
        except ProviderError:
            logger.exception('Error when processing video')
        except OSError:
            logger.exception('OS error when processing video')
        except UnicodeError:
            logger.exception('Character encoding error when processing video')
        except api.KnowitException as e:
            logger.error(e)

        if options.report and i % 20 == 19 and report:
            console.info('Unknown values so far:')
            console.info(dumps(report, options, vars(options)))

    if options.report:
        if report:
            console.info('Knowit %s found unknown values:', __version__)
            console.info(dumps(report, options, vars(options)))
            console.info('Please report them at %s', __url__)
        else:
            console.info('Knowit %s knows everything. :-)', __version__)


if __name__ == '__main__':
    main(sys.argv[1:])




import json
import logging
import re
from decimal import Decimal
from logging import NullHandler, getLogger
from subprocess import check_output

from knowit.core import Property
from knowit.properties import (
    AudioCodec,
    Basic,
    Duration,
    Language,
    Quantity,
    VideoCodec,
    VideoDimensions,
    YesNo,
)
from knowit.provider import (
    MalformedFileError,
    Provider,
)
from knowit.rules import (
    AudioChannelsRule,
    ClosedCaptionRule,
    HearingImpairedRule,
    LanguageRule,
    ResolutionRule,
)
from knowit.rules.general import GuessTitleRule
from knowit.serializer import get_json_encoder
from knowit.units import units
from knowit.utils import define_candidate, detect_os

logger = getLogger(__name__)
logger.addHandler(NullHandler())

WARN_MSG = r'''
=========================================================================================
mkvmerge not found on your system or could not be loaded.
Visit https://mkvtoolnix.download to download it.
If you still have problems, please check if the downloaded version matches your system.
To load mkvmerge from a specific location, please define the location as follow:
  knowit --mkvmerge /usr/local/mkvmerge/bin <video_path>
  knowit --mkvmerge /usr/local/mkvmerge/bin/ffprobe <video_path>
  knowit --mkvmerge "C:\Program Files\mkvmerge" <video_path>
  knowit --mkvmerge C:\Software\mkvmerge.exe <video_path>
=========================================================================================
'''


class MkvMergeExecutor:
    """Executor that knows how to execute mkvmerge."""

    version_re = re.compile(r'\bv(?P<version>[^\b\s]+)')
    locations = {
        'unix': ('/usr/local/mkvmerge/lib', '/usr/local/mkvmerge/bin', '__PATH__'),
        'windows': ('__PATH__', ),
        'macos': ('__PATH__', ),
    }

    def __init__(self, location, version):
        """Initialize the object."""
        self.location = location
        self.version = version

    def extract_info(self, filename):
        """Extract media info."""
        json_dump = self._execute(filename)
        return json.loads(json_dump) if json_dump else {}

    def _execute(self, filename):
        raise NotImplementedError

    @classmethod
    def _get_version(cls, output):
        match = cls.version_re.search(output)
        if match:
            version = match.groupdict()['version']
            return version

    @classmethod
    def get_executor_instance(cls, suggested_path=None):
        """Return executor instance."""
        os_family = detect_os()
        logger.debug('Detected os: %s', os_family)
        for exec_cls in (MkvMergeCliExecutor, ):
            executor = exec_cls.create(os_family, suggested_path)
            if executor:
                return executor


class MkvMergeCliExecutor(MkvMergeExecutor):
    """Executor that uses mkvmerge cli."""

    names = {
        'unix': ('mkvmerge', ),
        'windows': ('mkvmerge.exe', ),
        'macos': ('mkvmerge', ),
    }

    def _execute(self, filename):
        return check_output([self.location, '-i', '-F', 'json', filename]).decode()

    @classmethod
    def create(cls, os_family=None, suggested_path=None):
        """Create the executor instance."""
        for candidate in define_candidate(cls.locations, cls.names, os_family, suggested_path):
            try:
                output = check_output([candidate, '--version']).decode()
                version = cls._get_version(output)
                if version:
                    logger.debug('MkvMerge cli detected: %s v%s', candidate, version)
                    return MkvMergeCliExecutor(candidate, version.split('.'))
            except OSError:
                pass


class MkvMergeProvider(Provider):
    """MkvMerge Provider."""

    def __init__(self, config, suggested_path=None, *args, **kwargs):
        """Init method."""
        super().__init__(config, {
            'general': {
                'title': Property('title', description='media title'),
                'duration': Duration('duration', resolution=Decimal('0.000001'), description='media duration'),
            },
            'video': {
                'id': Basic('number', data_type=int, description='video track number'),
                'name': Property('name', description='video track name'),
                'language': Language('language_ietf', 'language', description='video language'),
                'width': VideoDimensions('display_dimensions', dimension='width'),
                'height': VideoDimensions('display_dimensions', dimension='height'),
                'scan_type': YesNo('interlaced', yes='Interlaced', no='Progressive', default='Progressive',
                                   config=config, config_key='ScanType',
                                   description='video scan type'),
                'resolution': None,  # populated with ResolutionRule
                # 'bit_depth', Property('bit_depth', Integer('video bit depth')),
                'codec': VideoCodec(config, 'codec_id', description='video codec'),
                'forced': YesNo('forced_track', hide_value=False, description='video track forced'),
                'default': YesNo('default_track', hide_value=False, description='video track default'),
                'enabled': YesNo('enabled_track', hide_value=True, description='video track enabled'),
            },
            'audio': {
                'id': Basic('number', data_type=int, description='audio track number'),
                'name': Property('name', description='audio track name'),
                'language': Language('language_ietf', 'language', description='audio language'),
                'codec': AudioCodec(config, 'codec_id', description='audio codec'),
                'channels_count': Basic('audio_channels', data_type=int, description='audio channels count'),
                'channels': None,  # populated with AudioChannelsRule
                'sampling_rate': Quantity('audio_sampling_frequency', unit=units.Hz, description='audio sampling rate'),
                'forced': YesNo('forced_track', hide_value=False, description='audio track forced'),
                'default': YesNo('default_track', hide_value=False, description='audio track default'),
                'enabled': YesNo('enabled_track', hide_value=True, description='audio track enabled'),
            },
            'subtitle': {
                'id': Basic('number', data_type=int, description='subtitle track number'),
                'name': Property('name', description='subtitle track name'),
                'language': Language('language_ietf', 'language', description='subtitle language'),
                'hearing_impaired': None,  # populated with HearingImpairedRule
                'closed_caption': None,  # populated with ClosedCaptionRule
                'forced': YesNo('forced_track', hide_value=False, description='subtitle track forced'),
                'default': YesNo('default_track', hide_value=False, description='subtitle track default'),
                'enabled': YesNo('enabled_track', hide_value=True, description='subtitle track enabled'),
            },
        }, {
            'video': {
                'guessed': GuessTitleRule('guessed properties', private=True),
                'language': LanguageRule('video language', override=True),
                'resolution': ResolutionRule('video resolution'),
            },
            'audio': {
                'guessed': GuessTitleRule('guessed properties', private=True),
                'language': LanguageRule('audio language', override=True),
                'channels': AudioChannelsRule('audio channels'),
            },
            'subtitle': {
                'guessed': GuessTitleRule('guessed properties', private=True),
                'language': LanguageRule('subtitle language', override=True),
                'hearing_impaired': HearingImpairedRule('subtitle hearing impaired', override=True),
                'closed_caption': ClosedCaptionRule('closed caption', override=True),
            }
        })
        self.executor = MkvMergeExecutor.get_executor_instance(suggested_path)

    def accepts(self, video_path):
        """Accept Matroska videos when mkvmerge is available."""
        if self.executor is None:
            logger.warning(WARN_MSG)
            self.executor = False

        return self.executor and video_path.lower().endswith(('.mkv', '.mka', '.mks'))

    @classmethod
    def extract_info(cls, video_path):
        """Extract info from the video."""
        return json.loads(check_output(['mkvmerge', '-i', '-F', video_path]).decode())

    def describe(self, video_path, context):
        """Return video metadata."""
        data = self.executor.extract_info(video_path)

        def debug_data():
            """Debug data."""
            return json.dumps(data, cls=get_json_encoder(context), indent=4, ensure_ascii=False)

        context['debug_data'] = debug_data

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('Video %r scanned using mkvmerge %r has raw data:\n%s',
                         video_path, self.executor.location, debug_data())

        def merge_properties(target: dict):
            """Merge properties sub properties into the target container."""
            return {**{k: v for k, v in target.items() if k != 'properties'}, **target.get('properties', {})}

        general_track = merge_properties(data.get('container', {}))
        video_tracks = []
        audio_tracks = []
        subtitle_tracks = []
        for track in data.get('tracks'):
            track_type = track.get('type')
            merged = merge_properties(track)
            if track_type == 'video':
                video_tracks.append(merged)
            elif track_type == 'audio':
                audio_tracks.append(merged)
            elif track_type == 'subtitles':
                subtitle_tracks.append(merged)

        result = self._describe_tracks(video_path, general_track, video_tracks, audio_tracks, subtitle_tracks, context)

        if not result:
            raise MalformedFileError

        result['provider'] = {
            'name': 'mkvmerge',
            'version': self.version
        }

        return result

    @property
    def version(self):
        """Return mkvmerge version information."""
        if not self.executor:
            return {}
        version = '.'.join(map(str, self.executor.version))

        return {self.executor.location: f'v{version}'}




import os
import typing
from logging import NullHandler, getLogger

import knowit.config
from knowit.core import Property, Rule
from knowit.properties import Quantity
from knowit.units import units

logger = getLogger(__name__)
logger.addHandler(NullHandler())


size_property = Quantity('size', unit=units.byte, description='media size')

PropertyMap = typing.Mapping[str, Property]
PropertyConfig = typing.Mapping[str, PropertyMap]

RuleMap = typing.Mapping[str, Rule]
RuleConfig = typing.Mapping[str, RuleMap]


class Provider:
    """Base class for all providers."""

    min_fps = 10
    max_fps = 200

    def __init__(
            self,
            config: knowit.config.Config,
            mapping: PropertyConfig,
            rules: typing.Optional[RuleConfig] = None,
    ):
        """Init method."""
        self.config = config
        self.mapping = mapping
        self.rules = rules or {}

    def accepts(self, target):
        """Whether or not the video is supported by this provider."""
        raise NotImplementedError

    def describe(self, target, context):
        """Read video metadata information."""
        raise NotImplementedError

    def _describe_tracks(self, video_path, general_track, video_tracks, audio_tracks, subtitle_tracks, context):
        logger.debug('Handling general track')
        props = self._describe_track(general_track, 'general', context)

        if 'path' not in props:
            props['path'] = video_path
        if 'container' not in props:
            props['container'] = os.path.splitext(video_path)[1][1:]
        if 'size' not in props and os.path.isfile(video_path):
            props['size'] = size_property.handle(os.path.getsize(video_path), context)

        for track_type, tracks, in (('video', video_tracks),
                                    ('audio', audio_tracks),
                                    ('subtitle', subtitle_tracks)):
            results = []
            for track in tracks or []:
                logger.debug('Handling %s track', track_type)
                t = self._validate_track(track_type, self._describe_track(track, track_type, context))
                if t:
                    results.append(t)

            if results:
                props[track_type] = results

        return props

    @classmethod
    def _validate_track(cls, track_type, track):
        if track_type != 'video' or 'frame_rate' not in track:
            return track

        frame_rate = track['frame_rate']
        try:
            frame_rate = frame_rate.magnitude
        except AttributeError:
            pass

        if cls.min_fps < frame_rate < cls.max_fps:
            return track

    def _describe_track(self, track, track_type, context):
        """Describe track to a dict.

        :param track:
        :param track_type:
        :rtype: dict
        """
        props = {}
        pv_props = {}
        for name, prop in self.mapping[track_type].items():
            if not prop:
                # placeholder to be populated by rules. It keeps the order
                props[name] = None
                continue

            value = prop.extract_value(track, context)
            if value is not None:
                which = props if not prop.private else pv_props
                which[name] = value

        for name, rule in self.rules.get(track_type, {}).items():
            if props.get(name) is not None and not rule.override:
                logger.debug('Skipping rule %s since property is already present: %r', name, props[name])
                continue

            value = rule.execute(props, pv_props, context)
            if value is not None:
                which = props if not rule.private else pv_props
                which[name] = value
            elif name in props and (not rule.override or props[name] is None):
                del props[name]

        return props

    @property
    def version(self):
        """Return provider version information."""
        raise NotImplementedError


class ProviderError(Exception):
    """Base class for provider exceptions."""

    pass


class MalformedFileError(ProviderError):
    """Malformed File error."""

    pass


class UnsupportedFileFormatError(ProviderError):
    """Unsupported File Format error."""

    pass



from decimal import Decimal

from knowit.core import Rule


class ResolutionRule(Rule):
    """Resolution rule."""

    standard_resolutions = (
        480,
        720,
        1080,
        2160,
        4320,
    )
    uncommon_resolutions = (
        240,
        288,
        360,
        576,
    )
    resolutions = list(sorted(standard_resolutions + uncommon_resolutions))
    square = 4. / 3
    wide = 16. / 9

    def execute(self, props, pv_props, context):
        """Return the resolution for the video.

        The resolution is based on a widescreen TV (16:9)
        1920x800 will be considered 1080p since the TV will use 1920x1080 with vertical black bars
        1426x1080 is considered 1080p since the TV will use 1920x1080 with horizontal black bars

        The calculation considers the display aspect ratio and the pixel aspect ratio (not only width and height).
        The upper resolution is selected if there's no perfect match with the following list of resolutions:
            240, 288, 360, 480, 576, 720, 1080, 2160, 4320
        If no interlaced information is available, resolution will be considered Progressive.
        """
        width = props.get('width')
        height = props.get('height')
        if not width or not height:
            return

        try:
            width = width.magnitude
            height = height.magnitude
        except AttributeError:
            pass

        dar = props.get('aspect_ratio', Decimal(width) / height)
        par = props.get('pixel_aspect_ratio', 1)
        scan_type = props.get('scan_type', 'p')[0].lower()

        # selected DAR must be between 4:3 and 16:9
        selected_dar = max(min(dar, self.wide), self.square)

        # mod-16
        stretched_width = int(round(width * par / 16)) * 16

        # mod-8
        calculated_height = int(round(stretched_width / selected_dar / 8)) * 8

        selected_resolution = None
        for r in reversed(self.resolutions):
            if r < calculated_height:
                break

            selected_resolution = r

        if selected_resolution:
            return f'{selected_resolution}{scan_type}'

        msg = f'{width}x{height} - scan_type: {scan_type}, aspect_ratio: {dar}, pixel_aspect_ratio: {par}'
        self.report(msg, context)



import os
import traceback
import typing

from knowit import __version__
from knowit.config import Config
from knowit.provider import Provider
from .providers import (
    EnzymeProvider,
    FFmpegProvider,
    MediaInfoProvider,
    MkvMergeProvider,
)

_provider_map = {
    'mediainfo': MediaInfoProvider,
    'ffmpeg': FFmpegProvider,
    'mkvmerge': MkvMergeProvider,
    'enzyme': EnzymeProvider,
}

provider_names = _provider_map.keys()

available_providers: typing.Dict[str, Provider] = {}


class KnowitException(Exception):
    """Exception raised when knowit encounters an internal error."""


def initialize(context: typing.Optional[typing.Mapping] = None) -> None:
    """Initialize knowit."""
    if not available_providers:
        context = context or {}
        config = Config.build(context.get('config'))
        for name, provider_cls in _provider_map.items():
            general_config = getattr(config, 'general', {})
            mapping = context.get(name) or general_config.get(name)
            available_providers[name] = provider_cls(config, mapping)


def know(
        video_path: typing.Union[str, os.PathLike],
        context: typing.Optional[typing.MutableMapping] = None
) -> typing.Mapping:
    """Return a mapping of video metadata."""
    video_path = os.fspath(video_path)

    try:
        context = context or {}
        context.setdefault('profile', 'default')
        initialize(context)

        for name, provider in available_providers.items():
            if name != (context.get('provider') or name):
                continue

            if provider.accepts(video_path):
                result = provider.describe(video_path, context)
                if result:
                    return result

        return {}
    except Exception:
        raise KnowitException(debug_info(context=context, exc_info=True))


def dependencies(context: typing.Optional[typing.Mapping] = None) -> typing.Mapping:
    """Return all dependencies detected by knowit."""
    deps = {}
    try:
        initialize(context)
        for name, provider_cls in _provider_map.items():
            if name in available_providers:
                deps[name] = available_providers[name].version
            else:
                deps[name] = {}
    except Exception:
        pass

    return deps


def _centered(value: str) -> str:
    value = value[-52:]
    return f'| {value:^53} |'


def debug_info(
        context: typing.Optional[typing.MutableMapping] = None,
        exc_info: bool = False,
) -> str:
    lines = [
        '+-------------------------------------------------------+',
        _centered(f'KnowIt {__version__}'),
        '+-------------------------------------------------------+'
    ]

    first = True
    for key, info in dependencies(context).items():
        if not first:
            lines.append(_centered(''))
        first = False

        for k, v in info.items():
            lines.append(_centered(k))
            lines.append(_centered(v))

    if context:
        debug_data = context.pop('debug_data', None)

        lines.append('+-------------------------------------------------------+')
        for k, v in context.items():
            if v:
                lines.append(_centered(f'{k}: {v}'))

        if debug_data:
            lines.append('+-------------------------------------------------------+')
            lines.append(debug_data())

    if exc_info:
        lines.append('+-------------------------------------------------------+')
        lines.append(traceback.format_exc())

    lines.append('+-------------------------------------------------------+')
    lines.append(_centered('Please report any bug or feature request at'))
    lines.append(_centered('https://github.com/ratoaq2/knowit/issues.'))
    lines.append('+-------------------------------------------------------+')

    return '\n'.join(lines)




from knowit.core import Configurable


class SubtitleFormat(Configurable[str]):
    """Subtitle Format property."""

    @classmethod
    def _extract_key(cls, value) -> str:
        key = str(value).upper()
        if key.startswith('S_'):
            key = key[2:]

        return key.split('/')[-1]



import typing

from knowit.core import Configurable, Property


class BitRateMode(Configurable[str]):
    """Bit Rate mode property."""


class AudioCompression(Configurable[str]):
    """Audio Compression property."""


class AudioProfile(Configurable[str]):
    """Audio profile property."""


class AudioChannels(Property[int]):
    """Audio Channels property."""

    ignored = {
        'object based',  # Dolby Atmos
    }

    def handle(self, value: typing.Union[int, str], context: typing.MutableMapping) -> typing.Optional[int]:
        """Handle audio channels."""
        if isinstance(value, int):
            return value

        if value.lower() not in self.ignored:
            try:
                return int(value)
            except ValueError:
                self.report(value, context)
        return None


class AudioCodec(Configurable[str]):
    """Audio codec property."""

    @classmethod
    def _extract_key(cls, value) -> str:
        key = str(value).upper()
        if key.startswith('A_'):
            key = key[2:]

        # only the first part of the word. E.g.: 'AAC LC' => 'AAC'
        return key.split(' ')[0]

    @classmethod
    def _extract_fallback_key(cls, value, key) -> typing.Optional[str]:
        if '/' in key:
            return key.split('/')[0]
        else:
            return None



import os
import typing
from logging import NullHandler, getLogger

from pkg_resources import resource_stream
import yaml

from knowit.serializer import get_yaml_loader

logger = getLogger(__name__)
logger.addHandler(NullHandler())


class _Value(typing.NamedTuple):
    code: str
    default: str
    human: str
    technical: str


_valid_aliases = _Value._fields


class Config:
    """Application config class."""

    @classmethod
    def build(cls, path: typing.Optional[typing.Union[str, os.PathLike]] = None) -> 'Config':
        """Build config instance."""
        loader = get_yaml_loader()
        with resource_stream('knowit', 'defaults.yml') as stream:
            cfgs = [yaml.load(stream, Loader=loader)]

        if path:
            with open(path, 'rb') as stream:
                cfgs.append(yaml.load(stream, Loader=loader))

        profiles_data = {}
        for cfg in cfgs:
            if 'profiles' in cfg:
                profiles_data.update(cfg['profiles'])

        knowledge_data = {}
        for cfg in cfgs:
            if 'knowledge' in cfg:
                knowledge_data.update(cfg['knowledge'])

        data: typing.Dict[str, typing.MutableMapping] = {'general': {}}
        for class_name, data_map in knowledge_data.items():
            data.setdefault(class_name, {})
            for code, detection_values in data_map.items():
                alias_map = (profiles_data.get(class_name) or {}).get(code) or {}
                alias_map.setdefault('code', code)
                alias_map.setdefault('default', alias_map['code'])
                alias_map.setdefault('human', alias_map['default'])
                alias_map.setdefault('technical', alias_map['human'])
                value = _Value(**{k: v for k, v in alias_map.items() if k in _valid_aliases})
                for detection_value in detection_values:
                    data[class_name][str(detection_value)] = value

        config = Config()
        config.__dict__ = data
        return config



from logging import NullHandler, getLogger

from trakit.api import trakit

from knowit.core import Rule

logger = getLogger(__name__)
logger.addHandler(NullHandler())


class GuessTitleRule(Rule):
    """Guess properties from track title."""

    def execute(self, props, pv_props, context):
        """Language detection using name."""
        if 'name' in props:
            language = props.get('language')
            options = {'expected_language': language} if language else {}
            guessed = trakit(props['name'], options)
            if guessed:
                return guessed


class LanguageRule(Rule):
    """Language rules."""

    def execute(self, props, pv_props, context):
        """Language detection using name."""
        if 'guessed' not in pv_props:
            return

        guess = pv_props['guessed']
        if 'language' in guess:
            return guess['language']



import typing


class NullRegistry:
    """A NullRegistry that masquerades as a pint.UnitRegistry."""

    def __init__(self):
        """Initialize a null registry."""

    def __getattr__(self, item: typing.Any) -> int:
        """Return a Scalar 1 to simulate a unit."""
        return 1

    def __bool__(self):
        """Return False since a NullRegistry is not a pint.UnitRegistry."""
        return False

    def define(self, *args, **kwargs):
        """Pretend to add unit to the registry."""


def _build_unit_registry():
    try:
        import pint

        registry = pint.UnitRegistry()
        registry.define('FPS = 1 * hertz')

        pint.set_application_registry(registry)
        return registry
    except ModuleNotFoundError:
        pass

    return NullRegistry()


units = _build_unit_registry()



import datetime
import json
import re
import typing
from datetime import timedelta
from decimal import Decimal

import babelfish
import yaml
from yaml.composer import Composer
from yaml.constructor import SafeConstructor
from yaml.parser import Parser
from yaml.reader import Reader
from yaml.resolver import Resolver as DefaultResolver
from yaml.scanner import Scanner

from knowit.units import units
from knowit.utils import round_decimal


def format_property(profile: str, o):
    """Convert properties to string."""
    if isinstance(o, timedelta):
        return format_duration(o, profile)

    if isinstance(o, babelfish.language.Language):
        return format_language(o, profile)

    if hasattr(o, 'units'):
        return format_quantity(o, profile)

    return str(o)


def get_json_encoder(context):
    """Return json encoder that handles all needed object types."""
    class StringEncoder(json.JSONEncoder):
        """String json encoder."""

        def default(self, o):
            return format_property(context['profile'], o)

    return StringEncoder


def get_yaml_dumper(context):
    """Return yaml dumper that handles all needed object types."""
    class CustomDumper(yaml.SafeDumper):
        """Custom YAML Dumper."""

        def default_representer(self, data):
            """Convert data to string."""
            if isinstance(data, int):
                return self.represent_int(data)
            return self.represent_str(str(data))

        def default_language_representer(self, data):
            """Convert language to string."""
            return self.represent_str(format_language(data, context['profile']))

        def default_quantity_representer(self, data):
            """Convert quantity to string."""
            return self.default_representer(format_quantity(data, context['profile']))

        def default_duration_representer(self, data):
            """Convert quantity to string."""
            return self.default_representer(format_duration(data, context['profile']))

    CustomDumper.add_representer(babelfish.Language, CustomDumper.default_language_representer)
    CustomDumper.add_representer(timedelta, CustomDumper.default_duration_representer)
    CustomDumper.add_representer(units.Quantity, CustomDumper.default_quantity_representer)
    CustomDumper.add_representer(Decimal, CustomDumper.default_representer)

    return CustomDumper


def get_yaml_loader(constructors=None):
    """Return a yaml loader that handles sequences as python lists."""
    constructors = constructors or {}
    yaml_implicit_resolvers = dict(DefaultResolver.yaml_implicit_resolvers)

    class Resolver(DefaultResolver):
        """Custom YAML Resolver."""

    Resolver.yaml_implicit_resolvers.clear()
    for ch, vs in yaml_implicit_resolvers.items():
        Resolver.yaml_implicit_resolvers.setdefault(ch, []).extend(
            (tag, regexp) for tag, regexp in vs
            if not tag.endswith('float')
        )
    Resolver.add_implicit_resolver(  # regex copied from yaml source
        '!decimal',
        re.compile(r'''^(?:
            [-+]?(?:[0-9][0-9_]*)\.[0-9_]*(?:[eE][-+][0-9]+)?
            |\.[0-9_]+(?:[eE][-+][0-9]+)?
            |[-+]?[0-9][0-9_]*(?::[0-9]?[0-9])+\.[0-9_]*
            |[-+]?\.(?:inf|Inf|INF)
            |\.(?:nan|NaN|NAN)
        )$''', re.VERBOSE),
        list('-+0123456789.')
    )

    class CustomLoader(Reader, Scanner, Parser, Composer, SafeConstructor, Resolver):
        """Custom YAML Loader."""

        def __init__(self, stream):
            Reader.__init__(self, stream)
            Scanner.__init__(self)
            Parser.__init__(self)
            Composer.__init__(self)
            SafeConstructor.__init__(self)
            Resolver.__init__(self)

    CustomLoader.add_constructor('tag:yaml.org,2002:seq', yaml.Loader.construct_python_tuple)
    for tag, constructor in constructors.items():
        CustomLoader.add_constructor(tag, constructor)

    def decimal_constructor(loader, node):
        value = loader.construct_scalar(node)
        return Decimal(value)

    CustomLoader.add_constructor('!decimal', decimal_constructor)

    return CustomLoader


def format_duration(
        duration: datetime.timedelta,
        profile='default',
) -> typing.Union[str, Decimal]:
    if profile == 'technical':
        return str(duration)

    seconds = duration.total_seconds()
    if profile == 'code':
        return round_decimal(
            Decimal((duration.days * 86400 + duration.seconds) * 10 ** 6 + duration.microseconds) / 10**6, min_digits=1
        )

    hours = int(seconds // 3600)
    seconds = seconds - (hours * 3600)
    minutes = int(seconds // 60)
    seconds = int(seconds - (minutes * 60))
    if profile == 'human':
        if hours > 0:
            return f'{hours} hours {minutes:02d} minutes { seconds:02d} seconds'
        if minutes > 0:
            return f'{minutes} minutes {seconds:02d} seconds'
        return f'{seconds} seconds'

    return f'{hours}:{minutes:02d}:{seconds:02d}'


def format_language(
        language: babelfish.language.Language,
        profile: str = 'default',
) -> str:
    if profile in ('default', 'human'):
        return str(language.name)

    return str(language)


def format_quantity(
        quantity,
        profile='default',
) -> str:
    """Human friendly format."""
    if profile == 'code':
        return quantity.magnitude

    unit = quantity.units
    if unit != 'bit':
        technical = profile == 'technical'
        if unit == 'hertz':
            return _format_quantity(quantity.magnitude, unit='Hz', binary=technical, precision=3 if technical else 1)

        root_unit = quantity.to_root_units().units
        if root_unit == 'bit':
            return _format_quantity(quantity.magnitude, binary=technical, precision=3 if technical else 2)
        if root_unit == 'bit / second':
            return _format_quantity(quantity.magnitude, unit='bps', binary=technical, precision=3 if technical else 1)

    return str(quantity)


def _format_quantity(
        num,
        unit: str = 'B',
        binary: bool = False,
        precision: int = 2,
) -> str:
    if binary:
        factor = 1024
        affix = 'i'
    else:
        factor = 1000
        affix = ''
    for prefix in ('', 'K', 'M', 'G', 'T', 'P', 'E', 'Z'):
        if abs(num) < factor:
            break
        num /= factor
    else:
        prefix = 'Y'

    return f'{num:3.{precision}f} {prefix}{affix}{unit}'


YAMLLoader = get_yaml_loader()




import json
import logging
import re
from logging import NullHandler, getLogger
from subprocess import check_output

from knowit import VIDEO_EXTENSIONS
from knowit.core import Property
from knowit.properties import (
    AudioChannels,
    AudioCodec,
    AudioProfile,
    Basic,
    Duration,
    Language,
    Quantity,
    Ratio,
    ScanType,
    SubtitleFormat,
    VideoCodec,
    VideoProfile,
    VideoProfileLevel,
    YesNo,
)
from knowit.provider import (
    MalformedFileError,
    Provider,
)
from knowit.rules import (
    AudioChannelsRule,
    ClosedCaptionRule,
    HearingImpairedRule,
    LanguageRule,
    ResolutionRule,
)
from knowit.rules.general import GuessTitleRule
from knowit.serializer import get_json_encoder
from knowit.units import units
from knowit.utils import (
    define_candidate,
    detect_os,
)

logger = getLogger(__name__)
logger.addHandler(NullHandler())


WARN_MSG = r'''
=========================================================================================
FFmpeg (ffprobe) not found on your system or could not be loaded.
Visit https://ffmpeg.org/download.html to download it.
If you still have problems, please check if the downloaded version matches your system.
To load FFmpeg (ffprobe) from a specific location, please define the location as follow:
  knowit --ffmpeg /usr/local/ffmpeg/bin <video_path>
  knowit --ffmpeg /usr/local/ffmpeg/bin/ffprobe <video_path>
  knowit --ffmpeg "C:\Program Files\FFmpeg" <video_path>
  knowit --ffmpeg C:\Software\ffprobe.exe <video_path>
=========================================================================================
'''


class FFmpegExecutor:
    """Executor that knows how to execute media info: using ctypes or cli."""

    version_re = re.compile(r'\bversion\s+(?P<version>[^\b\s]+)')
    locations = {
        'unix': ('/usr/local/ffmpeg/lib', '/usr/local/ffmpeg/bin', '__PATH__'),
        'windows': ('__PATH__', ),
        'macos': ('__PATH__', ),
    }

    def __init__(self, location, version):
        """Initialize the object."""
        self.location = location
        self.version = version

    def extract_info(self, filename):
        """Extract media info."""
        json_dump = self._execute(filename)
        return json.loads(json_dump) if json_dump else {}

    def _execute(self, filename):
        raise NotImplementedError

    @classmethod
    def _get_version(cls, output):
        match = cls.version_re.search(output)
        if match:
            version = match.groupdict()['version']
            return version

    @classmethod
    def get_executor_instance(cls, suggested_path=None):
        """Return executor instance."""
        os_family = detect_os()
        logger.debug('Detected os: %s', os_family)
        for exec_cls in (FFmpegCliExecutor, ):
            executor = exec_cls.create(os_family, suggested_path)
            if executor:
                return executor


class FFmpegCliExecutor(FFmpegExecutor):
    """Executor that uses FFmpeg (ffprobe) cli."""

    names = {
        'unix': ('ffprobe', ),
        'windows': ('ffprobe.exe', ),
        'macos': ('ffprobe', ),
    }

    def _execute(self, filename):
        return check_output([self.location, '-v', 'quiet', '-print_format', 'json',
                            '-show_format', '-show_streams', '-sexagesimal', filename]).decode()

    @classmethod
    def create(cls, os_family=None, suggested_path=None):
        """Create the executor instance."""
        for candidate in define_candidate(cls.locations, cls.names, os_family, suggested_path):
            try:
                output = check_output([candidate, '-version']).decode()
                version = cls._get_version(output)
                if version:
                    logger.debug('FFmpeg cli detected: %s v%s', candidate, version)
                    return FFmpegCliExecutor(candidate, version.split('.'))
            except OSError:
                pass


class FFmpegProvider(Provider):
    """FFmpeg provider."""

    def __init__(self, config, suggested_path=None):
        """Init method."""
        super().__init__(config, {
            'general': {
                'title': Property('tags.title', description='media title'),
                'path': Property('filename', description='media path'),
                'duration': Duration('duration', description='media duration'),
                'size': Quantity('size', unit=units.byte, description='media size'),
                'bit_rate': Quantity('bit_rate', unit=units.bps, description='media bit rate'),
            },
            'video': {
                'id': Basic('index', data_type=int, allow_fallback=True, description='video track number'),
                'name': Property('tags.title', description='video track name'),
                'language': Language('tags.language', description='video language'),
                'duration': Duration('duration', 'tags.duration', description='video duration'),
                'width': Quantity('width', unit=units.pixel),
                'height': Quantity('height', unit=units.pixel),
                'scan_type': ScanType(config, 'field_order', default='Progressive', description='video scan type'),
                'aspect_ratio': Ratio('display_aspect_ratio', description='display aspect ratio'),
                'pixel_aspect_ratio': Ratio('sample_aspect_ratio', description='pixel aspect ratio'),
                'resolution': None,  # populated with ResolutionRule
                'frame_rate': Ratio('r_frame_rate', unit=units.FPS, description='video frame rate'),
                # frame_rate_mode
                'bit_rate': Quantity('bit_rate', 'tags.bps', unit=units.bps, description='video bit rate'),
                'bit_depth': Quantity('bits_per_raw_sample', unit=units.bit, description='video bit depth'),
                'codec': VideoCodec(config, 'codec_name', description='video codec'),
                'profile': VideoProfile(config, 'profile', description='video codec profile'),
                'profile_level': VideoProfileLevel(config, 'level', description='video codec profile level'),
                # 'profile_tier': VideoProfileTier(config, 'codec_profile', description='video codec profile tier'),
                'forced': YesNo('disposition.forced', hide_value=False, description='video track forced'),
                'default': YesNo('disposition.default', hide_value=False, description='video track default'),
            },
            'audio': {
                'id': Basic('index', data_type=int, allow_fallback=True, description='audio track number'),
                'name': Property('tags.title', description='audio track name'),
                'language': Language('tags.language', description='audio language'),
                'duration': Duration('duration', 'tags.duration', description='audio duration'),
                'codec': AudioCodec(config, 'profile', 'codec_name', description='audio codec'),
                'profile': AudioProfile(config, 'profile', description='audio codec profile'),
                'channels_count': AudioChannels('channels', description='audio channels count'),
                'channels': None,  # populated with AudioChannelsRule
                'bit_depth': Quantity('bits_per_raw_sample', unit=units.bit, description='audio bit depth'),
                'bit_rate': Quantity('bit_rate', 'tags.bps', unit=units.bps, description='audio bit rate'),
                'sampling_rate': Quantity('sample_rate', unit=units.Hz, description='audio sampling rate'),
                'forced': YesNo('disposition.forced', hide_value=False, description='audio track forced'),
                'default': YesNo('disposition.default', hide_value=False, description='audio track default'),
            },
            'subtitle': {
                'id': Basic('index', data_type=int, allow_fallback=True, description='subtitle track number'),
                'name': Property('tags.title', description='subtitle track name'),
                'language': Language('tags.language', description='subtitle language'),
                'hearing_impaired': YesNo('disposition.hearing_impaired',
                                          hide_value=False, description='subtitle hearing impaired'),
                'closed_caption': None,  # populated with ClosedCaptionRule
                'format': SubtitleFormat(config, 'codec_name', description='subtitle format'),
                'forced': YesNo('disposition.forced', hide_value=False, description='subtitle track forced'),
                'default': YesNo('disposition.default', hide_value=False, description='subtitle track default'),
            },
        }, {
            'video': {
                'guessed': GuessTitleRule('guessed properties', private=True),
                'language': LanguageRule('video language', override=True),
                'resolution': ResolutionRule('video resolution'),
            },
            'audio': {
                'guessed': GuessTitleRule('guessed properties', private=True),
                'language': LanguageRule('audio language', override=True),
                'channels': AudioChannelsRule('audio channels'),
            },
            'subtitle': {
                'guessed': GuessTitleRule('guessed properties', private=True),
                'language': LanguageRule('subtitle language', override=True),
                'hearing_impaired': HearingImpairedRule('subtitle hearing impaired', override=True),
                'closed_caption': ClosedCaptionRule('closed caption', override=True),
            },
        })
        self.executor = FFmpegExecutor.get_executor_instance(suggested_path)

    def accepts(self, video_path):
        """Accept any video when FFprobe is available."""
        if self.executor is None:
            logger.warning(WARN_MSG)
            self.executor = False

        return self.executor and video_path.lower().endswith(VIDEO_EXTENSIONS)

    def describe(self, video_path, context):
        """Return video metadata."""
        data = self.executor.extract_info(video_path)

        def debug_data():
            """Debug data."""
            return json.dumps(data, cls=get_json_encoder(context), indent=4, ensure_ascii=False)

        context['debug_data'] = debug_data

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('Video %r scanned using ffmpeg %r has raw data:\n%s',
                         video_path, self.executor.location, debug_data())

        general_track = data.get('format') or {}
        if 'tags' in general_track:
            general_track['tags'] = {k.lower(): v for k, v in general_track['tags'].items()}

        video_tracks = []
        audio_tracks = []
        subtitle_tracks = []
        for track in data.get('streams'):
            track_type = track.get('codec_type')
            if track_type == 'video':
                video_tracks.append(track)
            elif track_type == 'audio':
                audio_tracks.append(track)
            elif track_type == 'subtitle':
                subtitle_tracks.append(track)

        result = self._describe_tracks(video_path, general_track, video_tracks, audio_tracks, subtitle_tracks, context)
        if not result:
            raise MalformedFileError

        result['provider'] = self.executor.location
        result['provider'] = {
            'name': 'ffmpeg',
            'version': self.version
        }

        return result

    @property
    def version(self):
        """Return ffmpeg version information."""
        if not self.executor:
            return {}
        version = '.'.join(map(str, self.executor.version))

        return {self.executor.location: f'v{version}'}



import re

from knowit.core import Rule


class ClosedCaptionRule(Rule):
    """Closed caption rule."""

    cc_re = re.compile(r'(\bcc\d\b)', re.IGNORECASE)

    def execute(self, props, pv_props, context):
        """Execute closed caption rule."""
        if '_closed_caption' in pv_props and self.cc_re.search(pv_props['_closed_caption']):
            return True

        if 'guessed' in pv_props:
            guessed = pv_props['guessed']
            return guessed.get('closed_caption')


class HearingImpairedRule(Rule):
    """Hearing Impaired rule."""

    def execute(self, props, pv_props, context):
        """Hearing Impaired."""
        if 'guessed' in pv_props:
            guessed = pv_props['guessed']
            return guessed.get('hearing_impaired')



"""Know your media files better."""
from importlib import metadata

__title__ = metadata.metadata(__package__)['name']
__version__ = metadata.version(__package__)
__short_version__ = '.'.join(__version__.split('.')[:2])
__author__ = metadata.metadata(__package__)['author']
__license__ = metadata.metadata(__package__)['license']
__url__ = metadata.metadata(__package__)['home_page']

del metadata

#: Video extensions
VIDEO_EXTENSIONS = ('.3g2', '.3gp', '.3gp2', '.3gpp', '.60d', '.ajp', '.asf', '.asx', '.avchd', '.avi', '.bik',
                    '.bix', '.box', '.cam', '.dat', '.divx', '.dmf', '.dv', '.dvr-ms', '.evo', '.flc', '.fli',
                    '.flic', '.flv', '.flx', '.gvi', '.gvp', '.h264', '.m1v', '.m2p', '.m2ts', '.m2v', '.m4e',
                    '.m4v', '.mjp', '.mjpeg', '.mjpg', '.mk3d', '.mkv', '.moov', '.mov', '.movhd', '.movie', '.movx',
                    '.mp4', '.mpe', '.mpeg', '.mpg', '.mpv', '.mpv2', '.mxf', '.nsv', '.nut', '.ogg', '.ogm', '.ogv',
                    '.omf', '.ps', '.qt', '.ram', '.rm', '.rmvb', '.swf', '.ts', '.vfw', '.vid', '.video', '.viv',
                    '.vivo', '.vob', '.vro', '.webm', '.wm', '.wmv', '.wmx', '.wrap', '.wvx', '.wx', '.x264', '.xvid')

from knowit.api import KnowitException, know



"""Provider package."""

from knowit.providers.enzyme import EnzymeProvider
from knowit.providers.ffmpeg import FFmpegProvider
from knowit.providers.mediainfo import MediaInfoProvider
from knowit.providers.mkvmerge import MkvMergeProvider



import typing
from decimal import Decimal
from logging import NullHandler, getLogger

from knowit.core import Rule

logger = getLogger(__name__)
logger.addHandler(NullHandler())


class AtmosRule(Rule):
    """Atmos rule."""

    def __init__(self, config: typing.Mapping[str, typing.Mapping], name: str,
                 **kwargs):
        """Initialize an Atmos rule."""
        super().__init__(name, **kwargs)
        self.audio_codecs = getattr(config, 'AudioCodec')

    def execute(self, props, pv_props, context):
        """Execute the rule against properties."""
        profile = context.get('profile') or 'default'
        format_commercial = pv_props.get('format_commercial')
        if 'codec' in props and format_commercial and 'atmos' in format_commercial.lower():
            props['codec'] = [props['codec'],
                              getattr(self.audio_codecs['ATMOS'], profile)]


class AudioChannelsRule(Rule):
    """Audio Channel rule."""

    mapping = {
        1: '1.0',
        2: '2.0',
        6: '5.1',
        8: '7.1',
    }

    def execute(self, props, pv_props, context):
        """Execute the rule against properties."""
        count = props.get('channels_count')
        if count is None:
            return

        channels = self.mapping.get(count) if isinstance(count, int) else None
        positions = pv_props.get('channel_positions') or []
        positions = positions if isinstance(positions, list) else [positions]
        candidate = 0
        for position in positions:
            if not position:
                continue

            c = Decimal('0.0')
            for i in position.split('/'):
                try:
                    c += Decimal(i)
                except ValueError:
                    logger.debug('Invalid %s: %s', self.description, i)
                    pass

            c_count = int(c) + int(round((c - int(c)) * 10))
            if c_count == count:
                return str(c)

            candidate = max(candidate, c)

        if channels:
            return channels

        if candidate:
            return candidate

        self.report(positions, context)


class DtsHdRule(Rule):
    """DTS-HD rule."""

    def __init__(self, config: typing.Mapping[str, typing.Mapping], name: str,
                 **kwargs):
        """Initialize a DTS-HD Rule."""
        super().__init__(name, **kwargs)
        self.audio_codecs = getattr(config, 'AudioCodec')
        self.audio_profiles = getattr(config, 'AudioProfile')

    @classmethod
    def _redefine(cls, props, name, index):
        actual = props.get(name)
        if isinstance(actual, list):
            value = actual[index]
            if value is None:
                del props[name]
            else:
                props[name] = value

    def execute(self, props, pv_props, context):
        """Execute the rule against properties."""
        profile = context.get('profile') or 'default'

        if props.get('codec') == getattr(self.audio_codecs['DTS'],
                                         profile) and props.get('profile') in (
                getattr(self.audio_profiles['MA'], profile),
                getattr(self.audio_profiles['HRA'], profile)):
            props['codec'] = getattr(self.audio_codecs['DTS-HD'], profile)



from knowit.properties.audio import (
    AudioChannels,
    AudioCodec,
    AudioCompression,
    AudioProfile,
    BitRateMode,
)
from knowit.properties.general import (
    Basic,
    Duration,
    Language,
    Quantity,
    YesNo,
)
from knowit.properties.subtitle import (
    SubtitleFormat,
)
from knowit.properties.video import (
    Ratio,
    ScanType,
    VideoCodec,
    VideoDimensions,
    VideoEncoder,
    VideoHdrFormat,
    VideoProfile,
    VideoProfileLevel,
    VideoProfileTier,
)



import re
import typing
from datetime import timedelta
from decimal import Decimal, InvalidOperation

import babelfish

from knowit.core import Configurable, Property
from knowit.utils import round_decimal

T = typing.TypeVar('T')


class Basic(Property[T]):
    """Basic property to handle int, Decimal and other basic types."""

    def __init__(self, *args: str, data_type: typing.Type,
                 processor: typing.Optional[typing.Callable[[T], T]] = None,
                 allow_fallback: bool = False, **kwargs):
        """Init method."""
        super().__init__(*args, **kwargs)
        self.data_type = data_type
        self.processor = processor or (lambda x: x)
        self.allow_fallback = allow_fallback

    def handle(self, value, context: typing.MutableMapping):
        """Handle value."""
        if isinstance(value, self.data_type):
            return self.processor(value)

        try:
            return self.processor(self.data_type(value))
        except ValueError:
            if not self.allow_fallback:
                self.report(value, context)


class Duration(Property[timedelta]):
    """Duration property."""

    duration_re = re.compile(r'(?P<hours>\d{1,2}):'
                             r'(?P<minutes>\d{1,2}):'
                             r'(?P<seconds>\d{1,2})(?:\.'
                             r'(?P<milliseconds>\d{3})'
                             r'(?P<microseconds>\d{3})?\d*)?')

    def __init__(self, *args: str, resolution: typing.Union[int, Decimal] = 1, **kwargs):
        """Initialize a Duration."""
        super().__init__(*args, **kwargs)
        self.resolution = resolution

    def handle(self, value, context: typing.MutableMapping):
        """Return duration as timedelta."""
        if isinstance(value, timedelta):
            return value
        elif isinstance(value, int):
            return timedelta(milliseconds=int(value * self.resolution))
        try:
            return timedelta(
                milliseconds=int(Decimal(value) * self.resolution))
        except (ValueError, InvalidOperation):
            pass

        match = self.duration_re.match(value)
        if not match:
            self.report(value, context)
            return None

        params = {
            key: int(value)
            for key, value in match.groupdict().items()
            if value
        }
        return timedelta(**params)


class Language(Property[babelfish.Language]):
    """Language property."""

    def handle(self, value, context: typing.MutableMapping):
        """Handle languages."""
        try:
            if len(value) == 3:
                return babelfish.Language.fromalpha3b(value)

            return babelfish.Language.fromietf(value)
        except (babelfish.Error, ValueError):
            pass

        try:
            return babelfish.Language.fromname(value)
        except babelfish.Error:
            pass

        self.report(value, context)
        return babelfish.Language('und')


class Quantity(Property):
    """Quantity is a property with unit."""

    def __init__(self, *args: str, unit, data_type=int, **kwargs):
        """Init method."""
        super().__init__(*args, **kwargs)
        self.unit = unit
        self.data_type = data_type

    def handle(self, value, context):
        """Handle value with unit."""
        if not isinstance(value, self.data_type):
            try:
                value = self.data_type(value)
            except ValueError:
                self.report(value, context)
                return
        if isinstance(value, Decimal):
            value = round_decimal(value, min_digits=1, max_digits=3)

        return value if context.get('no_units') else value * self.unit


class YesNo(Configurable[str]):
    """Yes or No handler."""

    yes_values = ('yes', 'true', '1')

    def __init__(self, *args: str, yes=True, no=False, hide_value=None,
                 config: typing.Optional[
                     typing.Mapping[str, typing.Mapping]] = None,
                 config_key: typing.Optional[str] = None,
                 **kwargs):
        """Init method."""
        super().__init__(config or {}, config_key=config_key, *args, **kwargs)
        self.yes = yes
        self.no = no
        self.hide_value = hide_value

    def handle(self, value, context):
        """Handle boolean values."""
        result = self.yes if str(value).lower() in self.yes_values else self.no
        if result == self.hide_value:
            return None

        return super().handle(result, context) if self.mapping else result



import re
import typing
from decimal import Decimal

from knowit.core import Configurable
from knowit.core import Property
from knowit.utils import round_decimal


class VideoCodec(Configurable[str]):
    """Video Codec handler."""

    @classmethod
    def _extract_key(cls, value) -> str:
        key = value.upper().split('/')[-1]
        if key.startswith('V_'):
            key = key[2:]

        return key.split(' ')[-1]


class VideoDimensions(Property[int]):
    """Dimensions property."""

    def __init__(self, *args: str, dimension='width' or 'height', **kwargs):
        """Initialize the object."""
        super().__init__(*args, **kwargs)
        self.dimension = dimension

    dimensions_re = re.compile(r'(?P<width>\d+)x(?P<height>\d+)')

    def handle(self, value, context) -> typing.Optional[int]:
        """Handle ratio."""
        match = self.dimensions_re.match(value)
        if match:
            match_dict = match.groupdict()
            try:
                value = match_dict[self.dimension]
            except KeyError:
                pass
            else:
                return int(value)

        self.report(value, context)
        return None


class VideoEncoder(Configurable):
    """Video Encoder property."""


class VideoHdrFormat(Configurable):
    """Video HDR Format property."""


class VideoProfile(Configurable[str]):
    """Video Profile property."""

    @classmethod
    def _extract_key(cls, value) -> str:
        return value.upper().split('@')[0]


class VideoProfileLevel(Configurable[str]):
    """Video Profile Level property."""

    @classmethod
    def _extract_key(cls, value) -> typing.Union[str, bool]:
        values = str(value).upper().split('@')
        if len(values) > 1:
            value = values[1]
            return value

        # There's no level, so don't warn or report it
        return False


class VideoProfileTier(Configurable[str]):
    """Video Profile Tier property."""

    @classmethod
    def _extract_key(cls, value) -> typing.Union[str, bool]:
        values = str(value).upper().split('@')
        if len(values) > 2:
            return values[2]

        # There's no tier, so don't warn or report it
        return False


class Ratio(Property[Decimal]):
    """Ratio property."""

    def __init__(self, *args: str, unit=None, **kwargs):
        """Initialize the object."""
        super().__init__(*args, **kwargs)
        self.unit = unit

    ratio_re = re.compile(r'(?P<width>\d+)[:/](?P<height>\d+)')

    def handle(self, value, context) -> typing.Optional[Decimal]:
        """Handle ratio."""
        match = self.ratio_re.match(value)
        if match:
            width, height = match.groups()
            if (width, height) == ('0', '1'):  # identity
                return Decimal('1.0')

            if height:
                result = round_decimal(Decimal(width) / Decimal(height), min_digits=1, max_digits=3)
                if self.unit:
                    result *= self.unit

                return result

        self.report(value, context)
        return None


class ScanType(Configurable[str]):
    """Scan Type property."""




from knowit.rules.audio import AtmosRule
from knowit.rules.audio import AudioChannelsRule
from knowit.rules.audio import DtsHdRule
from knowit.rules.general import LanguageRule
from knowit.rules.subtitle import ClosedCaptionRule
from knowit.rules.subtitle import HearingImpairedRule
from knowit.rules.video import ResolutionRule



import typing
from logging import NullHandler, getLogger

logger = getLogger(__name__)
logger.addHandler(NullHandler())

T = typing.TypeVar('T')

_visible_chars_table = dict.fromkeys(range(32))


def _is_unknown(value: typing.Any) -> bool:
    return isinstance(value, str) and (not value or value.lower() == 'unknown')


class Reportable(typing.Generic[T]):
    """Reportable abstract class."""

    def __init__(
            self,
            *args: str,
            description: typing.Optional[str] = None,
            reportable: bool = True,
    ):
        """Initialize the object."""
        self.names = args
        self._description = description
        self.reportable = reportable

    @property
    def description(self) -> str:
        """Rule description."""
        return self._description or '|'.join(self.names)

    def report(self, value: typing.Union[str, T], context: typing.MutableMapping) -> None:
        """Report unknown value."""
        if not value or not self.reportable:
            return

        if 'report' in context:
            report_map = context['report'].setdefault(self.description, {})
            if value not in report_map:
                report_map[value] = context['path']
        logger.info('Invalid %s: %r', self.description, value)


class Property(Reportable[T]):
    """Property class."""

    def __init__(
            self,
            *args: str,
            default: typing.Optional[T] = None,
            private: bool = False,
            description: typing.Optional[str] = None,
            delimiter: str = ' / ',
            **kwargs,
    ):
        """Init method."""
        super().__init__(*args, description=description, **kwargs)
        self.default = default
        self.private = private
        # Used to detect duplicated values. e.g.: en / en or High@L4.0 / High@L4.0 or Progressive / Progressive
        self.delimiter = delimiter

    @classmethod
    def _extract_value(cls,
                       track: typing.Mapping,
                       name: str,
                       names: typing.List[str]):
        if len(names) == 2:
            parent_value = track.get(names[0], track.get(names[0].upper(), {}))
            return parent_value.get(names[1], parent_value.get(names[1].upper()))

        return track.get(name, track.get(name.upper()))

    def extract_value(
            self,
            track: typing.Mapping,
            context: typing.MutableMapping,
    ) -> typing.Optional[T]:
        """Extract the property value from a given track."""
        for name in self.names:
            names = name.split('.')
            value = self._extract_value(track, name, names)
            if value is None:
                if self.default is None:
                    continue

                value = self.default

            if isinstance(value, bytes):
                value = value.decode()

            if isinstance(value, str):
                value = value.translate(_visible_chars_table).strip()
                if _is_unknown(value):
                    continue
                value = self._deduplicate(value)

            result = self.handle(value, context)
            if result is not None and not _is_unknown(result):
                return result

        return None

    @classmethod
    def _deduplicate(cls, value: str) -> str:
        values = value.split(' / ')
        if len(values) == 2 and values[0] == values[1]:
            return values[0]
        return value

    def handle(self, value: T, context: typing.MutableMapping) -> typing.Optional[T]:
        """Return the value without any modification."""
        return value


class Configurable(Property[T]):
    """Configurable property where values are in a config mapping."""

    def __init__(self, config: typing.Mapping[str, typing.Mapping], *args: str,
                 config_key: typing.Optional[str] = None, **kwargs):
        """Init method."""
        super().__init__(*args, **kwargs)
        self.mapping = getattr(config, config_key or self.__class__.__name__) if config else {}

    @classmethod
    def _extract_key(cls, value: str) -> typing.Union[str, bool]:
        return value.upper()

    @classmethod
    def _extract_fallback_key(cls, value: str, key: str) -> typing.Optional[T]:
        return None

    def _lookup(
            self,
            key: str,
            context: typing.MutableMapping,
    ) -> typing.Union[T, None, bool]:
        result = self.mapping.get(key)
        if result is not None:
            result = getattr(result, context.get('profile') or 'default')
            return result if result != '__ignored__' else False
        return None

    def handle(self, value, context):
        """Return Variable or Constant."""
        key = self._extract_key(value)
        if key is False:
            return

        result = self._lookup(key, context)
        if result is False:
            return

        while not result and key:
            key = self._extract_fallback_key(value, key)
            result = self._lookup(key, context)
            if result is False:
                return

        if not result:
            self.report(value, context)

        return result


class MultiValue(Property):
    """Property with multiple values."""

    def __init__(self, prop: typing.Optional[Property] = None, delimiter='/', single=False,
                 handler=None, name=None, **kwargs):
        """Init method."""
        super().__init__(*(prop.names if prop else (name,)), **kwargs)
        self.prop = prop
        self.delimiter = delimiter
        self.single = single
        self.handler = handler

    def handle(
            self,
            value: str,
            context: typing.MutableMapping,
    ) -> typing.Union[T, typing.List[T]]:
        """Handle properties with multiple values."""
        if self.handler:
            call = self.handler
        elif self.prop:
            call = self.prop.handle
        else:
            raise NotImplementedError('No handler available')

        result = call(value, context)
        if result is not None:
            return result

        if isinstance(value, list):
            if len(value) == 1:
                values = self._split(value[0], self.delimiter)
            else:
                values = value
        else:
            values = self._split(value, self.delimiter)

        if values is None:
            return call(values, context)
        if len(values) > 1 and not self.single:
            results = [call(item, context) if not _is_unknown(item) else None for item in values]
            results = [r for r in results if r is not None]
            if results:
                return results
        return call(values[0], context)

    @classmethod
    def _split(
            cls,
            value: typing.Optional[T],
            delimiter: str = '/',
    ) -> typing.Optional[typing.List[str]]:
        if value is None:
            return None

        return [x.strip() for x in str(value).split(delimiter)]


class Rule(Reportable[T]):
    """Rule abstract class."""

    def __init__(self, name: str, private=False, override=False, **kwargs):
        """Initialize the object."""
        super().__init__(name, **kwargs)
        self.private = private
        self.override = override

    def execute(self, props, pv_props, context: typing.Mapping):
        """How to execute a rule."""
        raise NotImplementedError



import os
import sys
import typing
from decimal import Decimal

from knowit import VIDEO_EXTENSIONS

if sys.version_info < (3, 8):
    OS_FAMILY = str
else:
    OS_FAMILY = typing.Literal['windows', 'macos', 'unix']

OPTION_MAP = typing.Dict[str, typing.Tuple[str]]


def recurse_paths(
        paths: typing.Union[str, typing.Iterable[str]]
) -> typing.List[str]:
    """Return a list of video files."""
    enc_paths = []

    if isinstance(paths, str):
        paths = [p.strip() for p in paths.split(',')] if ',' in paths else paths.split()

    for path in paths:
        if os.path.isfile(path):
            enc_paths.append(path)
        if os.path.isdir(path):
            for root, directories, filenames in os.walk(path):
                for filename in filenames:
                    if os.path.splitext(filename)[1] in VIDEO_EXTENSIONS:
                        full_path = os.path.join(root, filename)
                        enc_paths.append(full_path)

    # Lets remove any dupes since mediainfo is rather slow.
    unique_paths = dict.fromkeys(enc_paths)
    return list(unique_paths)


def to_dict(
        obj: typing.Any,
        classkey: typing.Optional[typing.Type] = None
) -> typing.Union[str, dict, list]:
    """Transform an object to dict."""
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, dict):
        data = {}
        for (k, v) in obj.items():
            data[k] = to_dict(v, classkey)
        return data
    elif hasattr(obj, '_ast'):
        return to_dict(obj._ast())
    elif hasattr(obj, '__iter__'):
        return [to_dict(v, classkey) for v in obj]
    elif hasattr(obj, '__dict__'):
        values = [(key, to_dict(value, classkey))
                  for key, value in obj.__dict__.items() if not callable(value) and not key.startswith('_')]
        data = {k: v for k, v in values if v is not None}
        if classkey is not None and hasattr(obj, '__class__'):
            data[classkey] = obj.__class__.__name__
        return data
    return obj


def detect_os() -> OS_FAMILY:
    """Detect os family: windows, macos or unix."""
    if os.name in ('nt', 'dos', 'os2', 'ce'):
        return 'windows'
    if sys.platform == 'darwin':
        return 'macos'
    return 'unix'


def define_candidate(
        locations: OPTION_MAP,
        names: OPTION_MAP,
        os_family: typing.Optional[OS_FAMILY] = None,
        suggested_path: typing.Optional[str] = None,
) -> typing.Generator[str, None, None]:
    """Select family-specific options and generate possible candidates."""
    os_family = os_family or detect_os()
    family_names = names[os_family]
    all_locations = (suggested_path, ) + locations[os_family]
    yield from build_candidates(all_locations, family_names)


def build_candidates(
        locations: typing.Iterable[typing.Optional[str]],
        names: typing.Iterable[str],
) -> typing.Generator[str, None, None]:
    """Build candidate names."""
    for location in locations:
        if not location:
            continue
        if location == '__PATH__':
            yield from build_path_candidates(names)
        elif os.path.isfile(location):
            yield location
        elif os.path.isdir(location):
            for name in names:
                cmd = os.path.join(location, name)
                if os.path.isfile(cmd):
                    yield cmd


def build_path_candidates(
    names: typing.Iterable[str],
    os_family: typing.Optional[OS_FAMILY] = None,
) -> typing.Generator[str, None, None]:
    """Build candidate names on environment PATH."""
    os_family = os_family or detect_os()
    if os_family != 'windows':
        yield from names
    else:
        paths = os.environ['PATH'].split(';')
        yield from (
            os.path.join(path, name)
            for path in paths
            for name in names
        )


def round_decimal(value: Decimal, min_digits=0, max_digits: typing.Optional[int] = None):
    exponent = value.normalize().as_tuple().exponent
    if exponent >= 0:
        return round(value, min_digits)

    decimal_places = abs(exponent)
    if decimal_places <= min_digits:
        return round(value, min_digits)
    if max_digits:
        return round(value, min(max_digits, decimal_places))
    return value




import json
import logging
from collections import defaultdict
from logging import NullHandler, getLogger
import enzyme

from knowit.core import Property
from knowit.properties import (
    AudioCodec,
    Basic,
    Duration,
    Language,
    Quantity,
    VideoCodec,
    YesNo,
)
from knowit.provider import (
    MalformedFileError,
    Provider,
)
from knowit.rules import (
    AudioChannelsRule,
    ClosedCaptionRule,
    HearingImpairedRule,
    LanguageRule,
    ResolutionRule,
)
from knowit.rules.general import GuessTitleRule
from knowit.serializer import get_json_encoder
from knowit.units import units
from knowit.utils import to_dict

logger = getLogger(__name__)
logger.addHandler(NullHandler())


class EnzymeProvider(Provider):
    """Enzyme Provider."""

    def __init__(self, config, *args, **kwargs):
        """Init method."""
        super().__init__(config, {
            'general': {
                'title': Property('title', description='media title'),
                'duration': Duration('duration', description='media duration'),
            },
            'video': {
                'id': Basic('number', data_type=int, description='video track number'),
                'name': Property('name', description='video track name'),
                'language': Language('language', description='video language'),
                'width': Quantity('width', unit=units.pixel),
                'height': Quantity('height', unit=units.pixel),
                'scan_type': YesNo('interlaced', yes='Interlaced', no='Progressive', default='Progressive',
                                   config=config, config_key='ScanType',
                                   description='video scan type'),
                'resolution': None,  # populated with ResolutionRule
                # 'bit_depth', Property('bit_depth', Integer('video bit depth')),
                'codec': VideoCodec(config, 'codec_id', description='video codec'),
                'forced': YesNo('forced', hide_value=False, description='video track forced'),
                'default': YesNo('default', hide_value=False, description='video track default'),
                'enabled': YesNo('enabled', hide_value=True, description='video track enabled'),
            },
            'audio': {
                'id': Basic('number', data_type=int, description='audio track number'),
                'name': Property('name', description='audio track name'),
                'language': Language('language', description='audio language'),
                'codec': AudioCodec(config, 'codec_id', description='audio codec'),
                'channels_count': Basic('channels', data_type=int, description='audio channels count'),
                'channels': None,  # populated with AudioChannelsRule
                'forced': YesNo('forced', hide_value=False, description='audio track forced'),
                'default': YesNo('default', hide_value=False, description='audio track default'),
                'enabled': YesNo('enabled', hide_value=True, description='audio track enabled'),
            },
            'subtitle': {
                'id': Basic('number', data_type=int, description='subtitle track number'),
                'name': Property('name', description='subtitle track name'),
                'language': Language('language', description='subtitle language'),
                'hearing_impaired': None,  # populated with HearingImpairedRule
                'closed_caption': None,  # populated with ClosedCaptionRule
                'forced': YesNo('forced', hide_value=False, description='subtitle track forced'),
                'default': YesNo('default', hide_value=False, description='subtitle track default'),
                'enabled': YesNo('enabled', hide_value=True, description='subtitle track enabled'),
            },
        }, {
            'video': {
                'guessed': GuessTitleRule('guessed properties', private=True),
                'language': LanguageRule('video language', override=True),
                'resolution': ResolutionRule('video resolution'),
            },
            'audio': {
                'guessed': GuessTitleRule('guessed properties', private=True),
                'language': LanguageRule('audio language', override=True),
                'channels': AudioChannelsRule('audio channels'),
            },
            'subtitle': {
                'guessed': GuessTitleRule('guessed properties', private=True),
                'language': LanguageRule('subtitle language', override=True),
                'hearing_impaired': HearingImpairedRule('subtitle hearing impaired', override=True),
                'closed_caption': ClosedCaptionRule('closed caption', override=True),
            }
        })

    def accepts(self, video_path):
        """Accept only MKV files."""
        return video_path.lower().endswith('.mkv')

    @classmethod
    def extract_info(cls, video_path):
        """Extract info from the video."""
        with open(video_path, 'rb') as f:
            return to_dict(enzyme.MKV(f))

    def describe(self, video_path, context):
        """Return video metadata."""
        try:
            data = defaultdict(dict)
            ff = self.extract_info(video_path)

            def debug_data():
                """Debug data."""
                return json.dumps(ff, cls=get_json_encoder(context), indent=4, ensure_ascii=False)
            context['debug_data'] = debug_data

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('Video %r scanned using enzyme %r has raw data:\n%s',
                             video_path, enzyme.__version__, debug_data)

            data.update(ff)
            if 'info' in data and data['info'] is None:
                return {}
        except enzyme.MalformedMKVError:  # pragma: no cover
            raise MalformedFileError

        if logger.level == logging.DEBUG:
            logger.debug('Video {video_path} scanned using Enzyme {version} has raw data:\n{data}',
                         video_path=video_path, version=enzyme.__version__,
                         data=json.dumps(data, cls=get_json_encoder(context), indent=4, ensure_ascii=False))

        result = self._describe_tracks(video_path, data.get('info', {}), data.get('video_tracks'),
                                       data.get('audio_tracks'), data.get('subtitle_tracks'), context)

        if not result:
            raise MalformedFileError

        result['provider'] = {
            'name': 'enzyme',
            'version': self.version
        }

        return result

    @property
    def version(self):
        """Return enzyme version information."""
        return {'enzyme': enzyme.__version__}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['knowit', 'knowit.properties', 'knowit.providers', 'knowit.rules']

package_data = \
{'': ['*']}

install_requires = \
['babelfish>=0.6.0,<0.7.0',
 'enzyme>=0.4.1,<0.5.0',
 'pint>=0.20.1,<0.21.0',
 'pymediainfo>=6.0.1,<7.0.0',
 'pyyaml>=6.0,<7.0',
 'trakit>=0.2.1,<0.3.0']

entry_points = \
{'console_scripts': ['knowit = knowit.__main__:main']}

setup_kwargs = {
    'name': 'knowit',
    'version': '0.5.2',
    'description': 'Know better your media files',
    'long_description': '# KnowIt\n\nKnow better your media files.\n\n[![Latest\nVersion](https://img.shields.io/pypi/v/knowit.svg)](https://pypi.python.org/pypi/knowit)\n\n[![tests](https://github.com/ratoaq2/knowit/actions/workflows/test.yml/badge.svg)](https://github.com/ratoaq2/knowit/actions/workflows/test.yml)\n\n[![License](https://img.shields.io/github/license/ratoaq2/knowit.svg)](https://github.com/ratoaq2/knowit/blob/master/LICENSE)\n\n  - Project page  \n    <https://github.com/ratoaq2/knowit>\n\n## Usage\n\n### CLI\n\nExtract information from a video file:\n\n    $ knowit /folder/Audio Samples/hd_dtsma_7.1.mkv\n    For: /folder/Audio Samples/hd_dtsma_7.1.mkv\n    Knowit 0.4.0 found:\n    {\n        "title": "7.1Ch DTS-HD MA - Speaker Mapping Test File",\n        "path": "/folder/Audio Samples/hd_dtsma_7.1.mkv",\n        "duration": "0:01:37",\n        "size": "40.77 MB",\n        "bit_rate": "3.3 Mbps",\n        "container": "mkv",\n        "video": [\n            {\n                "id": 1,\n                "duration": "0:01:37",\n                "width": "1920 pixel",\n                "height": "1080 pixel",\n                "scan_type": "Progressive",\n                "aspect_ratio": "1.778",\n                "pixel_aspect_ratio": "1.0",\n                "resolution": "1080p",\n                "frame_rate": "23.976 FPS",\n                "bit_depth": "8 bit",\n                "codec": "H.264",\n                "profile": "Main",\n                "profile_level": "4",\n                "media_type": "video/H264",\n                "default": true\n            }\n        ],\n        "audio": [\n            {\n                "id": 2,\n                "name": "7.1Ch DTS-HD MA",\n                "language": "English",\n                "duration": "0:01:37",\n                "codec": "DTS-HD",\n                "profile": "Master Audio",\n                "channels_count": 8,\n                "channels": "7.1",\n                "bit_depth": "24 bit",\n                "bit_rate_mode": "Variable",\n                "sampling_rate": "48.0 KHz",\n                "compression": "Lossless",\n                "default": true\n            }\n        ],\n        "provider": {\n            "name": "mediainfo",\n            "version": {\n                "pymediainfo": "5.0.3",\n                "libmediainfo.so.0": "v20.9"\n            }\n        }\n    }\n\nExtract information from a video file using ffmpeg:\n\n    $ knowit --provider ffmpeg /folder/Audio Samples/hd_dtsma_7.1.mkv\n    For: /folder/Audio Samples/hd_dtsma_7.1.mkv\n    Knowit 0.4.0 found:\n    {\n        "title": "7.1Ch DTS-HD MA - Speaker Mapping Test File",\n        "path": "/folder/Audio Samples/hd_dtsma_7.1.mkv",\n        "duration": "0:01:37",\n        "size": "40.77 MB",\n        "bit_rate": "3.3 Mbps",\n        "container": "mkv",\n        "video": [\n            {\n                "id": 0,\n                "width": "1920 pixel",\n                "height": "1080 pixel",\n                "scan_type": "Progressive",\n                "aspect_ratio": "1.778",\n                "pixel_aspect_ratio": "1.0",\n                "resolution": "1080p",\n                "frame_rate": "23.976 FPS",\n                "bit_depth": "8 bit",\n                "codec": "H.264",\n                "profile": "Main",\n                "default": true\n            }\n        ],\n        "audio": [\n            {\n                "id": 1,\n                "name": "7.1Ch DTS-HD MA",\n                "language": "English",\n                "codec": "DTS-HD",\n                "profile": "Master Audio",\n                "channels_count": 8,\n                "channels": "7.1",\n                "bit_depth": "24 bit",\n                "sampling_rate": "48.0 KHz",\n                "default": true\n            }\n        ],\n        "provider": {\n            "name": "ffmpeg",\n            "version": {\n                "ffprobe": "v4.2.4-1ubuntu0.1"\n            }\n        }\n    }\n\nUsing docker:\n\n    docker run -it --rm -v /folder:/folder knowit /folder/Audio Samples/hd_dtsma_7.1.mkv\n    For: /folder/Audio Samples/hd_dtsma_7.1.mkv\n    Knowit 0.4.0 found:\n    {\n        "title": "7.1Ch DTS-HD MA - Speaker Mapping Test File",\n        "path": "/folder/Audio Samples/hd_dtsma_7.1.mkv",\n        "duration": "0:01:37",\n        "size": "40.77 MB",\n        "bit_rate": "3.3 Mbps",\n        "container": "mkv",\n        "video": [\n            {\n                "id": 1,\n                "duration": "0:01:37",\n                "width": "1920 pixel",\n                "height": "1080 pixel",\n                "scan_type": "Progressive",\n                "aspect_ratio": "1.778",\n                "pixel_aspect_ratio": "1.0",\n                "resolution": "1080p",\n                "frame_rate": "23.976 FPS",\n                "bit_depth": "8 bit",\n                "codec": "H.264",\n                "profile": "Main",\n                "profile_level": "4",\n                "media_type": "video/H264",\n                "default": true\n            }\n        ],\n        "audio": [\n            {\n                "id": 2,\n                "name": "7.1Ch DTS-HD MA",\n                "language": "English",\n                "duration": "0:01:37",\n                "codec": "DTS-HD",\n                "profile": "Master Audio",\n                "channels_count": 8,\n                "channels": "7.1",\n                "bit_depth": "24 bit",\n                "bit_rate_mode": "Variable",\n                "sampling_rate": "48.0 KHz",\n                "compression": "Lossless",\n                "default": true\n            }\n        ],\n        "provider": {\n            "name": "mediainfo",\n            "version": {\n                "pymediainfo": "5.0.3",\n                "libmediainfo.so.0": "v20.9"\n            }\n        }\n    }\n\nAll available CLI options:\n\n    $ knowit --help\n    usage: knowit [-h] [-p PROVIDER] [--debug] [--report] [-y] [-N] [-P PROFILE] [--mediainfo MEDIAINFO] [--ffmpeg FFMPEG] [--mkvmerge MKVMERGE] [--version] [videopath [videopath ...]]\n    \n    positional arguments:\n      videopath             Path to the video to introspect\n    \n    optional arguments:\n      -h, --help            show this help message and exit\n    \n    Providers:\n      -p PROVIDER, --provider PROVIDER\n                            The provider to be used: mediainfo, ffmpeg, mkvmerge or enzyme.\n    \n    Output:\n      --debug               Print information for debugging knowit and for reporting bugs.\n      --report              Parse media and report all non-detected values\n      -y, --yaml            Display output in yaml format\n      -N, --no-units        Display output without units\n      -P PROFILE, --profile PROFILE\n                            Display values according to specified profile: code, default, human, technical\n    \n    Configuration:\n      --mediainfo MEDIAINFO\n                            The location to search for MediaInfo binaries\n      --ffmpeg FFMPEG       The location to search for ffprobe (FFmpeg) binaries\n      --mkvmerge MKVMERGE   The location to search for mkvmerge (MKVToolNix) binaries\n    \n    Information:\n      --version             Display knowit version.\n\n## Installation\n\nKnowIt can be installed as a regular python module by running:\n\n    $ [sudo] pip install knowit\n\nFor a better isolation with your system you should use a dedicated\nvirtualenv or install for your user only using the `--user` flag.\n\n## External dependencies\n\nKnowIt can use MediaInfo, ffprobe (FFmpeg) or mkvmerge (MKVToolNix)\n\nKnowIt supports MKV regardless if MediaInfo, FFmpeg or MKVToolNix are\ninstalled.\n\nMediaInfo, FFmpeg or MKVToolNix increases the number of supported\nformats and the number of extracted information.\n\nMediaInfo is the default provider. Visit their\n[website](http://mediaarea.net/MediaInfo) and install the proper package\nfor your system.\n\nffprobe (FFmpeg) can be downloaded\n[here](https://ffmpeg.org/download.html)\n\nmkvmerge (MKVToolNix) can be downloaded\n[here](https://mkvtoolnix.download/downloads.html)\n',
    'author': 'Rato',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ratoaq2/knowit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



import ctypes
import json
import os
import re
from ctypes import c_void_p, c_wchar_p
from decimal import Decimal
from logging import DEBUG, NullHandler, getLogger
from subprocess import CalledProcessError, check_output

from pymediainfo import MediaInfo
from pymediainfo import __version__ as pymediainfo_version

from knowit import VIDEO_EXTENSIONS
from knowit.core import MultiValue, Property
from knowit.properties import (
    AudioChannels,
    AudioCodec,
    AudioCompression,
    AudioProfile,
    Basic,
    BitRateMode,
    Duration,
    Language,
    Quantity,
    ScanType,
    SubtitleFormat,
    VideoCodec,
    VideoEncoder,
    VideoHdrFormat,
    VideoProfile,
    VideoProfileTier,
    YesNo,
)
from knowit.provider import (
    MalformedFileError,
    Provider,
)
from knowit.rules import (
    AtmosRule,
    AudioChannelsRule,
    ClosedCaptionRule,
    DtsHdRule,
    HearingImpairedRule,
    LanguageRule,
    ResolutionRule,
)
from knowit.rules.general import GuessTitleRule
from knowit.units import units
from knowit.utils import (
    define_candidate,
    detect_os, round_decimal,
)

logger = getLogger(__name__)
logger.addHandler(NullHandler())


WARN_MSG = r'''
=========================================================================================
MediaInfo not found on your system or could not be loaded.
Visit https://mediaarea.net/ to download it.
If you still have problems, please check if the downloaded version matches your system.
To load MediaInfo from a specific location, please define the location as follow:
  knowit --mediainfo /usr/local/mediainfo/lib <video_path>
  knowit --mediainfo /usr/local/mediainfo/bin <video_path>
  knowit --mediainfo "C:\Program Files\MediaInfo" <video_path>
  knowit --mediainfo C:\Software\MediaInfo.dll <video_path>
  knowit --mediainfo C:\Software\MediaInfo.exe <video_path>
  knowit --mediainfo /opt/mediainfo/libmediainfo.so <video_path>
  knowit --mediainfo /opt/mediainfo/libmediainfo.dylib <video_path>
=========================================================================================
'''


class MediaInfoExecutor:
    """Media info executable knows how to execute media info: using ctypes or cli."""

    version_re = re.compile(r'\bv(?P<version>\d+(?:\.\d+)+)\b')

    locations = {
        'unix': ('/usr/local/mediainfo/lib', '/usr/local/mediainfo/bin', '__PATH__'),
        'windows': ('C:\\Program Files\\MediaInfo', 'C:\\Program Files (x86)\\MediaInfo', '__PATH__'),
        'macos': ('__PATH__', ),
    }

    def __init__(self, location, version):
        """Initialize the object."""
        self.location = location
        self.version = version

    def extract_info(self, filename):
        """Extract media info."""
        return self._execute(filename)

    def _execute(self, filename):
        raise NotImplementedError

    @classmethod
    def _get_version(cls, output):
        match = cls.version_re.search(output)
        if match:
            version = tuple([int(v) for v in match.groupdict()['version'].split('.')])
            return version

    @classmethod
    def get_executor_instance(cls, suggested_path=None):
        """Return the executor instance."""
        os_family = detect_os()
        logger.debug('Detected os: %s', os_family)
        for exec_cls in (MediaInfoCTypesExecutor, MediaInfoCliExecutor):
            executor = exec_cls.create(os_family, suggested_path)
            if executor:
                return executor


class MediaInfoCliExecutor(MediaInfoExecutor):
    """Media info using cli."""

    names = {
        'unix': ('mediainfo', ),
        'windows': ('MediaInfo.exe', ),
        'macos': ('mediainfo', ),
    }

    def _execute(self, filename):
        data = check_output([self.location, '--Output=JSON', '--Full', filename]).decode()

        return json.loads(data) if data else {}

    @classmethod
    def _is_gui_exe(cls, candidate: str):
        if not candidate.endswith('MediaInfo.exe') or not os.path.isfile(candidate):
            return False

        try:
            shell32 = ctypes.WinDLL('shell32', use_last_error=True)  # type: ignore
            return bool(shell32.ExtractIconExW(candidate, 0, None, None, 1))
        except Exception:
            return False

    @classmethod
    def create(cls, os_family=None, suggested_path=None):
        """Create the executor instance."""
        for candidate in define_candidate(cls.locations, cls.names, os_family, suggested_path):
            if cls._is_gui_exe(candidate):
                continue

            try:
                output = check_output([candidate, '--version']).decode()
                version = cls._get_version(output)
                if version:
                    logger.debug('MediaInfo cli detected: %s', candidate)
                    return MediaInfoCliExecutor(candidate, version)
            except CalledProcessError as e:
                # old mediainfo returns non-zero exit code for mediainfo --version
                version = cls._get_version(e.output.decode())
                if version:
                    logger.debug('MediaInfo cli detected: %s', candidate)
                    return MediaInfoCliExecutor(candidate, version)
            except OSError:
                pass


class MediaInfoCTypesExecutor(MediaInfoExecutor):
    """Media info ctypes."""

    names = {
        'unix': ('libmediainfo.so.0', ),
        'windows': ('MediaInfo.dll', ),
        'macos': ('libmediainfo.0.dylib', 'libmediainfo.dylib'),
    }

    def _execute(self, filename):
        # Create a MediaInfo handle
        data = MediaInfo.parse(filename, library_file=self.location, output='JSON')

        return json.loads(data) if data else {}

    @classmethod
    def create(cls, os_family=None, suggested_path=None):
        """Create the executor instance."""
        for candidate in define_candidate(cls.locations, cls.names, os_family, suggested_path):
            if MediaInfo.can_parse(candidate):
                lib, handle, lib_version_str, lib_version = MediaInfo._get_library(candidate)
                lib.MediaInfo_Option.argtypes = [c_void_p, c_wchar_p, c_wchar_p]
                lib.MediaInfo_Option.restype = c_wchar_p
                version = MediaInfoExecutor._get_version(lib.MediaInfo_Option(None, "Info_Version", ""))

                logger.debug('MediaInfo library detected: %s (v%s)', candidate, '.'.join(map(str, version)))
                return MediaInfoCTypesExecutor(candidate, version)


class MediaInfoProvider(Provider):
    """Media Info provider."""

    executor = None

    def __init__(self, config, suggested_path):
        """Init method."""
        super().__init__(config, {
            'general': {
                'title': Property('Title', description='media title'),
                'path': Property('CompleteName', description='media path'),
                'duration': Duration('Duration', resolution=1000, description='media duration'),
                'size': Quantity('FileSize', unit=units.byte, description='media size'),
                'bit_rate': Quantity('OverallBitRate', unit=units.bps, description='media bit rate'),
            },
            'video': {
                'id': Basic('ID', data_type=int, allow_fallback=True, description='video track number'),
                'name': Property('Title', description='video track name'),
                'language': Language('Language', description='video language'),
                'duration': Duration('Duration', resolution=1000, description='video duration'),
                'size': Quantity('StreamSize', unit=units.byte, description='video stream size'),
                'width': Quantity('Width', unit=units.pixel),
                'height': Quantity('Height', unit=units.pixel),
                'scan_type': ScanType(config, 'ScanType', default='Progressive', description='video scan type'),
                'aspect_ratio': Basic('DisplayAspectRatio', data_type=Decimal,
                                      processor=lambda x: round_decimal(x, min_digits=1, max_digits=3),
                                      description='display aspect ratio'),
                'pixel_aspect_ratio': Basic('PixelAspectRatio', data_type=Decimal,
                                            processor=lambda x: round_decimal(x, min_digits=1, max_digits=3),
                                            description='pixel aspect ratio'),
                'resolution': None,  # populated with ResolutionRule
                'frame_rate': Quantity('FrameRate', unit=units.FPS, data_type=Decimal, description='video frame rate'),
                # frame_rate_mode
                'bit_rate': Quantity('BitRate', unit=units.bps, description='video bit rate'),
                'bit_depth': Quantity('BitDepth', unit=units.bit, description='video bit depth'),
                'codec': VideoCodec(config, 'CodecID', description='video codec'),
                'profile': VideoProfile(config, 'Format_Profile', description='video codec profile'),
                'profile_level': Property('Format_Level', description='video codec profile level'),
                'profile_tier': VideoProfileTier(config, 'Format_Tier', description='video codec profile tier'),
                'encoder': VideoEncoder(config, 'Encoded_Library_Name', description='video encoder'),
                'hdr_format': MultiValue(VideoHdrFormat(config, 'HDR_Format', description='video hdr format'),
                                         delimiter=' / '),
                'media_type': Property('InternetMediaType', description='video media type'),
                'forced': YesNo('Forced', hide_value=False, description='video track forced'),
                'default': YesNo('Default', hide_value=False, description='video track default'),
            },
            'audio': {
                'id': Basic('ID', data_type=int, allow_fallback=True, description='audio track number'),
                'name': Property('Title', description='audio track name'),
                'language': Language('Language', description='audio language'),
                'duration': Duration('Duration', resolution=1000, description='audio duration'),
                'size': Quantity('StreamSize', unit=units.byte, description='audio stream size'),
                'codec': MultiValue(AudioCodec(config, 'CodecID', description='audio codec')),
                'format_commercial': Property('Format_Commercial', private=True),
                'profile': MultiValue(AudioProfile(config, 'Format_Profile', 'Format_AdditionalFeatures',
                                                   description='audio codec profile'),
                                      delimiter=' / '),
                'channels_count': MultiValue(AudioChannels('Channels_Original', 'Channels',
                                                           description='audio channels count')),
                'channel_positions': MultiValue(name='ChannelPositions_String2', handler=(lambda x, *args: x),
                                                delimiter=' / ', private=True, description='audio channels position'),
                'channels': None,  # populated with AudioChannelsRule
                'bit_depth': Quantity('BitDepth', unit=units.bit, description='audio bit depth'),
                'bit_rate': MultiValue(Quantity('BitRate', unit=units.bps, description='audio bit rate')),
                'bit_rate_mode': MultiValue(BitRateMode(config, 'BitRate_Mode', description='audio bit rate mode')),
                'sampling_rate': MultiValue(Quantity('SamplingRate', unit=units.Hz, description='audio sampling rate')),
                'compression': MultiValue(AudioCompression(config, 'Compression_Mode',
                                                           description='audio compression')),
                'forced': YesNo('Forced', hide_value=False, description='audio track forced'),
                'default': YesNo('Default', hide_value=False, description='audio track default'),
            },
            'subtitle': {
                'id': Basic('ID', data_type=int, allow_fallback=True, description='subtitle track number'),
                'name': Property('Title', description='subtitle track name'),
                'language': Language('Language', description='subtitle language'),
                'hearing_impaired': None,  # populated with HearingImpairedRule
                '_closed_caption': Property('ClosedCaptionsPresent', private=True),
                'closed_caption': None,  # populated with ClosedCaptionRule
                'format': SubtitleFormat(config, 'CodecID', description='subtitle format'),
                'forced': YesNo('Forced', hide_value=False, description='subtitle track forced'),
                'default': YesNo('Default', hide_value=False, description='subtitle track default'),
            },
        }, {
            'video': {
                'guessed': GuessTitleRule('guessed properties', private=True),
                'language': LanguageRule('video language', override=True),
                'resolution': ResolutionRule('video resolution'),
            },
            'audio': {
                'guessed': GuessTitleRule('guessed properties', private=True),
                'language': LanguageRule('audio language', override=True),
                'channels': AudioChannelsRule('audio channels'),
                'atmos': AtmosRule(config, 'atmos rule', private=True),
                'dtshd': DtsHdRule(config, 'dts-hd rule', private=True),
            },
            'subtitle': {
                'guessed': GuessTitleRule('guessed properties', private=True),
                'language': LanguageRule('subtitle language', override=True),
                'hearing_impaired': HearingImpairedRule('subtitle hearing impaired', override=True),
                'closed_caption': ClosedCaptionRule('closed caption', override=True),
            }
        })
        self.executor = MediaInfoExecutor.get_executor_instance(suggested_path)

    def accepts(self, video_path):
        """Accept any video when MediaInfo is available."""
        if self.executor is None:
            logger.warning(WARN_MSG)
            self.executor = False

        return self.executor and video_path.lower().endswith(VIDEO_EXTENSIONS)

    def describe(self, video_path, context):
        """Return video metadata."""
        data = self.executor.extract_info(video_path)

        def debug_data():
            """Debug data."""
            return json.dumps(data, indent=4)

        context['debug_data'] = debug_data

        if logger.isEnabledFor(DEBUG):
            logger.debug('Video %r scanned using mediainfo %r has raw data:\n%s',
                         video_path, self.executor.location, debug_data())

        result = {}
        tracks = data.get('media', {}).get('track', [])
        if tracks:
            general_tracks = []
            video_tracks = []
            audio_tracks = []
            subtitle_tracks = []
            for track in tracks:
                track_type = track.get('@type')
                if track_type == 'General':
                    general_tracks.append(track)
                elif track_type == 'Video':
                    video_tracks.append(track)
                elif track_type == 'Audio':
                    audio_tracks.append(track)
                elif track_type == 'Text':
                    subtitle_tracks.append(track)

            result = self._describe_tracks(video_path, general_tracks[0] if general_tracks else {},
                                           video_tracks, audio_tracks, subtitle_tracks, context)
        if not result:
            raise MalformedFileError

        result['provider'] = {
            'name': 'mediainfo',
            'version': self.version
        }

        return result

    @property
    def version(self):
        """Return mediainfo version information."""
        versions = {'pymediainfo': pymediainfo_version}
        if self.executor:
            executor_version = '.'.join(map(str, self.executor.version))
            versions[self.executor.location] = f'v{executor_version}'
        return versions
